unit DBFloatEdit;

(*******************************************************************************

TFloatEdit/TSpinFloatEdit/TDBFloatEdit components by mirkes.de

TFloatEdit is a descendant of TCustomEdit which is able to edit and check
floating point values. You can set minimum and maximum values and the number of
digits to display.

TSpinFloatEdit is derived from TFloatEdit. It additionally implements increment
and decrement buttons. You can define a static increment value to use when
those buttons are clicked or compute the increment using an event handler.

TDBFloatEdit is a database aware version of TFloatEdit. It has initially been
written by Eddie Shipman.

-version:
version 06-15-2004, released june 15, 2004

-target:
the package has been tested and is supposed to work in
Borland Delphi 4, 6 and 7 (TDBFloatEdit has not been tested on Delphi 7).

see the end of this file for copyright, license, usage and history issues

*******************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Mask, DBCtrls, Db;

{$INCLUDE MPDELVER.INC}

type
  TCustomDBFloatEdit = class(TCustomMaskEdit)
  private
    FDataLink: TFieldDataLink;
    FCanvas: TControlCanvas;
    FAlignment: TAlignment;
    FFocused: Boolean;

    // ES
    FDigits      : Byte;
    FMin,FMax    : Extended;
    FDec         : Char;
    FErrorText   : string;
    FOldVal      : Extended;
    FTempValue   : Extended;
    // ES
    FOnEnterPress: TKeyPressEvent;
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetReadOnly: Boolean;
    function GetTextMargins: TPoint;
    procedure ResetMaxLength;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetFocused(Value: Boolean);
    procedure SetReadOnly(Value: Boolean);
    procedure UpdateData(Sender: TObject);
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure WMUndo(var Message: TMessage); message WM_UNDO;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;

    // ES
    procedure SetAlignment(AValue:TAlignment);
    procedure ShowError;
    function GetText: string;
    procedure SetValue(AValue : Extended);
    procedure SetMin(AValue : Extended);
    procedure SetMax(AValue : Extended);
    procedure SetDigits(AValue : Byte);
    function GetValue : Extended;
    // ES
    function GetAbout: string; virtual;
    procedure SetAbout(const Value: string);
  protected
    // ES
    property Digits: Byte read FDigits write SetDigits;
    property Text: string read GetText;
    property Min : Extended read FMin write SetMin;
    property Max : Extended read FMax write SetMax;
    property ErrorMessage :string read FErrorText write FErrorText;
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    // ES

    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    procedure Change; override;
    function EditCanModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Reset; override;

    procedure DoEnter;override;
    procedure DoExit;override;
    property Value    : Extended read GetValue write SetValue;
    property OnEnterPress: TKeyPressEvent read FOnEnterPress write
      FOnEnterPress;
    property About: string read GetAbout write SetAbout stored false;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ExecuteAction(Action: TBasicAction): Boolean; override;
    function UpdateAction(Action: TBasicAction): Boolean; override;
    property Field: TField read GetField;
  end;

  TDBFloatEdit = class(TCustomDBFloatEdit)
  published
    property About;
    property Digits;
    property Text;
    property Min;
    property Max;
    property ErrorMessage;
    property Alignment;
    property OnEnterPress;

    property Anchors;
    property AutoSelect;
    property AutoSize;
    property BiDiMode;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property Ctl3D;
    property DataField;
    property DataSource;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
{$IFDEF DELPHI5UP}
    property OnContextPopup;
{$ENDIF}
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

const
  NO_TEXT = '';
  SR_VER = '06-15-2004';
  SR_COPY =
    '1997-2004 Eddie Shipman and markus stephany, vcl[at]mirkes[dot]de, http://www.mirkes.de';


procedure Register;
begin
  RegisterComponents('mirkes.de', [TDBFloatEdit]);
end;

{ TCustomDBFloatEdit }

procedure TCustomDBFloatEdit.ResetMaxLength;
var
  F: TField;
begin
  if (MaxLength > 0) and Assigned(DataSource) and Assigned(DataSource.DataSet) then
  begin
    F := DataSource.DataSet.FindField(DataField);
    if Assigned(F) and (F.DataType in [ftString, ftWideString]) and (F.Size = MaxLength) then
      MaxLength := 0;
  end;
end;

constructor TCustomDBFloatEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;
  ControlStyle := ControlStyle + [csReplicatable];
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
  FDataLink.OnActiveChange := ActiveChange;
  // ES
  FDec := DecimalSeparator;
  FDigits := 1;
  FMin := 0;
  FMax := 100;//1E110;
  FErrorText := NO_TEXT;
  inherited Text := FloatToStrF(FMin,ffFixed,18,fDigits);
  // ES
end;

destructor TCustomDBFloatEdit.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TCustomDBFloatEdit.Loaded;
var
  F: Extended;
begin
  inherited Loaded;
  if FMin > FMax then
  begin
    F := FMax;
    FMax := FMin;
    FMin := F;
  end;
  Value := FTempValue;
  ResetMaxLength;
  if (csDesigning in ComponentState) then DataChange(Self);
end;

procedure TCustomDBFloatEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TCustomDBFloatEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TCustomDBFloatEdit.KeyPress(var Key: Char);
var
  ts: string;
  flTemp: Extended;
  intDigit: Integer;
begin
  case Key of
    #10, #13:
      begin
        Key := #0;
        Exit;
      end; {case: #10, #13}
    #32..#255:
      begin
        if not (Key in ['0'..'9', FDec, '-', 'e', 'E']) then
        begin
          inherited;
          Key := #0;
          Exit;
        end;

        // prfen, ob neuer text im erlaubten bereich liegt
        // seltext durch key ersetzen
        ts := Copy(Text, 1, SelStart) + Key + Copy(Text, SelStart + SelLength + 1,
          MaxInt);
        if ts = '-' then
        begin
          inherited;
          Key := #0;
          inherited Text := '-0';
          SelStart := 1;
          SelLength := 1;
          Exit;
        end;

        if ts = FDec then
        begin
          inherited;
          Key := #0;
          inherited Text := '0' + FDec + '0';
          SelStart := 2;
          SelLength := 1;
          Exit;
        end;

        if ts = '' then
        begin
          inherited;
          Key := #0;
          inherited Text := FloatToStrF(FMin, ffFixed, 18, FDigits);
          SelectAll;
          Exit;
        end;
        try
          flTemp := StrToFloat(ts);
        except
          inherited;
          Key := #0;
          Exit;
        end;

        if (*** (flTemp < FMin) or ***) // 1.5: removed, it would be impossible e.g.
        // to start typing the number '10' if the minimum value would be set to 2
        (flTemp > FMax) or
          ((Key = FDec) and (FDigits = 0)) then
        begin
          inherited;
          Key := #0;
          Exit;
        end;

        // anzahl der digits prfen (nicht bei exponentialschreibweise)
        if Pos('E', UpperCase(ts)) = 0 then
        begin
          intDigit := Pos(FDec, ts);
          if intDigit > 0 then
            if Length(ts) - intDigit > FDigits then
            begin
              inherited;
              Key := #0;
              Exit;
            end;
        end;

        inherited;
        FDataLink.Edit;
      end; {case: ^H, ^V, ^X, #32..#255}
    #27:
      begin
        Value := fOldVal;
        SelectAll;
        FDataLink.Reset;
        Key := #0;
        Exit;
      end; {case:#27}
    #8 {BS}, #22 {^V}, #24 {^X}:
      begin
        // Handle Backspace, Copy and Cut.
        // Just set the datalink into editmode and it handles
        // it from there.
        FDataLink.Edit;
      end;
  end; {case}
end;

function TCustomDBFloatEdit.EditCanModify: Boolean;
begin
  Result := FDataLink.Edit;
end;

procedure TCustomDBFloatEdit.SetAlignment(AValue:TAlignment);
begin
  if FAlignment<>AValue then
  begin
    FAlignment:=AValue;
    RecreateWnd;
  end;
end;

procedure TCustomDBFloatEdit.Reset;
begin
  FDataLink.Reset;
  SelectAll;
end;

procedure TCustomDBFloatEdit.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) and not IsMasked then Invalidate;
    FDataLink.Reset;
  end;
end;

procedure TCustomDBFloatEdit.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

function TCustomDBFloatEdit.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TCustomDBFloatEdit.SetDataSource(Value: TDataSource);
begin
  if not (FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
    FDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

function TCustomDBFloatEdit.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TCustomDBFloatEdit.SetDataField(const Value: string);
begin
  if not (csDesigning in ComponentState) then
    ResetMaxLength;
  FDataLink.FieldName := Value;
end;

function TCustomDBFloatEdit.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TCustomDBFloatEdit.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

function TCustomDBFloatEdit.GetField: TField;
begin
  Result := FDataLink.Field;
end;

procedure TCustomDBFloatEdit.ActiveChange(Sender: TObject);
begin
  ResetMaxLength;
end;

procedure TCustomDBFloatEdit.DataChange(Sender: TObject);
begin
  // ES
  if FDataLink.Field = nil then
    Value := 0.00
  else
    Value := FDataLink.Field.AsFloat;
  // ES
end;

procedure TCustomDBFloatEdit.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TCustomDBFloatEdit.UpdateData(Sender: TObject);
begin
  ValidateEdit;
  // FDataLink.Field.Text := Text;
  FDataLink.Field.AsFloat := Value;
end;

procedure TCustomDBFloatEdit.WMUndo(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TCustomDBFloatEdit.WMPaste(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TCustomDBFloatEdit.WMCut(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TCustomDBFloatEdit.CMEnter(var Message: TCMEnter);
begin
  SetFocused(True);
  inherited;
  if SysLocale.FarEast and FDataLink.CanModify then
    inherited ReadOnly := False;
end;

procedure TCustomDBFloatEdit.CMExit(var Message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;
  SetFocused(False);
  CheckCursor;
  DoExit;
end;

procedure TCustomDBFloatEdit.DoEnter;
begin
  if AutoSelect and not (csLButtonDown in ControlState) then
    SelectAll;
  FOldVal := GetValue;
  inherited;
end;

procedure TCustomDBFloatEdit.DoExit;
var
  ts: string;
  Res: Extended;
begin
  ts := Text;
  if ts = '' then
    ts := FloatToStrF(FMin, ffFixed, 18, fDigits);
  try
    Res := StrToFloat(ts);
    if (Res < FMin) or (Res > FMax) then
      raise Exception.Create(FErrorText);
  except
    ShowError;
    Value := FOldVal;
    SelectAll;
    SetFocus;
    Exit;
  end;
  // Ausgabe formatieren
  inherited Text := FloatToStrF(Res, ffFixed, 18, FDigits);
  Value := StrToFloat(Text);
  inherited;
end;

procedure TCustomDBFloatEdit.WMPaint(var Message: TWMPaint);
const
  AlignStyle : array[Boolean, TAlignment] of DWORD =
   ((WS_EX_LEFT, WS_EX_RIGHT, WS_EX_LEFT),
    (WS_EX_RIGHT, WS_EX_LEFT, WS_EX_LEFT));
var
  Left: Integer;
  Margins: TPoint;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
  AAlignment: TAlignment;
  ExStyle: DWORD;
begin
  AAlignment := FAlignment;
  if UseRightToLeftAlignment then ChangeBiDiModeAlignment(AAlignment);
  if ((AAlignment = taLeftJustify) or FFocused) and
    not (csPaintCopy in ControlState) then
  begin
    if SysLocale.MiddleEast and HandleAllocated and (IsRightToLeft) then
    begin { This keeps the right aligned text, right aligned }
      ExStyle := DWORD(GetWindowLong(Handle, GWL_EXSTYLE)) and (not WS_EX_RIGHT) and
        (not WS_EX_RTLREADING) and (not WS_EX_LEFTSCROLLBAR);
      if UseRightToLeftReading then ExStyle := ExStyle or WS_EX_RTLREADING;
      if UseRightToLeftScrollbar then ExStyle := ExStyle or WS_EX_LEFTSCROLLBAR;
      ExStyle := ExStyle or
        AlignStyle[UseRightToLeftAlignment, AAlignment];
      if DWORD(GetWindowLong(Handle, GWL_EXSTYLE)) <> ExStyle then
        SetWindowLong(Handle, GWL_EXSTYLE, ExStyle);
    end;
    inherited;
    Exit;
  end;
{ Since edit controls do not handle justification unless multi-line (and
  then only poorly) we will draw right and center justify manually unless
  the edit has the focus. }
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
      R := ClientRect;
      if not (NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      if not Enabled then
        Font.Color := clGrayText;
      if (csPaintCopy in ControlState) and (FDataLink.Field <> nil) then
      begin
        S := FDataLink.Field.DisplayText;
        case CharCase of
          ecUpperCase: S := AnsiUpperCase(S);
          ecLowerCase: S := AnsiLowerCase(S);
        end;
      end else
        S := EditText;
      if PasswordChar <> #0 then FillChar(S[1], Length(S), PasswordChar);
      Margins := GetTextMargins;
      case AAlignment of
        taLeftJustify: Left := Margins.X;
        taRightJustify: Left := ClientWidth - TextWidth(S) - Margins.X - 1;
      else
        Left := (ClientWidth - TextWidth(S)) div 2;
      end;
      if SysLocale.MiddleEast then UpdateTextFlags;
      TextRect(R, Left, Margins.Y, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TCustomDBFloatEdit.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(FDataLink);
end;

function TCustomDBFloatEdit.GetTextMargins: TPoint;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  if NewStyleControls then
  begin
    if BorderStyle = bsNone then I := 0 else
      if Ctl3D then I := 1 else I := 2;
    Result.X := SendMessage(Handle, EM_GETMARGINS, 0, 0) and $0000FFFF + I;
    Result.Y := I;
  end else
  begin
    if BorderStyle = bsNone then I := 0 else
    begin
      DC := GetDC(0);
      GetTextMetrics(DC, SysMetrics);
      SaveFont := SelectObject(DC, Font.Handle);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, SaveFont);
      ReleaseDC(0, DC);
      I := SysMetrics.tmHeight;
      if I > Metrics.tmHeight then I := Metrics.tmHeight;
      I := I div 4;
    end;
    Result.X := I;
    Result.Y := I;
  end;
end;

function TCustomDBFloatEdit.ExecuteAction(Action: TBasicAction): Boolean;
begin
  Result := inherited ExecuteAction(Action) or (FDataLink <> nil) and
    FDataLink.ExecuteAction(Action);
end;

function TCustomDBFloatEdit.UpdateAction(Action: TBasicAction): Boolean;
begin
  Result := inherited UpdateAction(Action) or (FDataLink <> nil) and
    FDataLink.UpdateAction(Action);
end;

procedure TCustomDBFloatEdit.SetValue(AValue : Extended);
var
  tmp: string;
begin
  if csLoading in ComponentState then
    FTempValue := AValue
  else
  begin
    if (AValue > FMax) then
    begin
      ShowError;
      AValue := FMax;
    end;
    if (AValue < FMin) then
    begin
      ShowError;
      AValue := FMin;
    end;
    tmp := FloatToStrF(AValue, ffFixed, 18, FDigits);
    inherited Text := tmp;
  end;
end;

function TCustomDBFloatEdit.GetValue : Extended;
var
  ts: string;
begin
  if csLoading in ComponentState then
    Result := FTempValue
  else
  begin
    ts := Text;
    if (ts = '-') or (ts = FDec) or (ts = '') or (ts = 'E') then
      ts := '0';
    try
      Result := StrToFloat(ts);
    except
      Result := FMin;
    end;
    if Result < FMin then
      Result := FMin
    else if Result > FMax then
      Result := FMax
  end;
end;

procedure TCustomDBFloatEdit.SetDigits(AValue : Byte);
begin
  if FDigits <> AValue then
  begin
    if AValue > 18
    then
      AValue := 18;
    FDigits := AValue;
    if FDataLink.Field = nil then
      Value := 0.00
    else
      Value := FDataLink.Field.AsFloat;
  end;
end;

procedure TCustomDBFloatEdit.SetMin(AValue : Extended);
begin
  if FMin <> AValue then
  begin // 06-15-2004
    if (AValue > FMax) and (not (csLoading in ComponentState)) then
      raise
        Exception.Create('Minimum value must not be greater than maximum value !');
    FMin := AValue;
    Value := Value;
  end;
end;

procedure TCustomDBFloatEdit.SetMax(AValue : Extended);
begin
  if FMax <> AValue then
  begin // 06-15-2004
    if (AValue < FMin) and (not (csLoading in ComponentState)) then
      raise
        Exception.Create('Maximum value must not be less than minimum value !');
    FMax := AValue;
    Value := Value;
  end;
end;

procedure TCustomDBFloatEdit.ShowError;
begin
  if FErrorText <> NO_TEXT
  then
    ShowMessage(FErrorText);
end;

function TCustomDBFloatEdit.GetText: string;
begin
  Result := inherited Text;
end;

function TCustomDBFloatEdit.GetAbout: string;
begin
  Result := self.ClassName + ' ' + SR_VER + ',' + SR_COPY
end;

procedure TCustomDBFloatEdit.SetAbout(const Value: string);
begin
  // no change
end;

(*******************************************************************************

-history:
version 06-15-2004, june 15, 2004
        changed SetMin and SetMax to allow temporary invalid values during
        component loading and check them after the component is loaded. thanks
        to Thorsten Brandt

        updated TDBFloatEdit to reflect the changes since version 1.2

        changed version numbering (now the version reflects the release date)

version 1.5, jan 27, 2003
        there was a problem when a minimum value of e.g. 2 was set, it wasn't
        possible to type the number '10'! thanks to Heino Kappler

version 1.4, jan 27, 2003
        some problems with return/escape keys fixed

        multiline style avoided on newer operating systems (this does not
        affect TSpinFloatEdit, it's always a multiline edit control)

        on win98/me/2k/xp+ escape/return now are only handled by the component
        itself if no cancel/default buttons are on the form. thanks to Moz

version 1.3c, nov 19, 2002
        entering numbers in scientific notation should now work better
        (reported by Scott Westfall and others)

version 1.3b, oct 27, 2001
        bugfix: OnExit had been called twice, thanks to Philippe Oppliger

        included a DB aware version of TFloatEdit 1.2 (TDBFloatEdit in
        dbfloatedit.pas), written by Eddie Shipman, many thanks!

version 1.3a, feb 23, 2001
        changed the license from GNU LGPL to the Clarified Artistic License

Version 1.3, jan 28, 2001
        implemented a base class TCustomFloatEdit which is the ancestor of
        TFloatEdit

        hopefully fixed a problem with different decimalseparators

        text property is no longer writable (use value instead)
        added a component TSplinFloatEdit (like TSpinEdit from Borland's sample
        vcl, much of the code is just copy&paste'd from the sources, not just
        fix increment/decrement, also computed - via event handler)

        credits to George Paraskevaidis

        credits to Borland for the TSpinEdit sample component

version 1.2
        added properties: Alignment, OnEnterPress
        credits to Giuseppe Mosca and Andrea Carli

version 1.1
        there was a bug in range checking; i hope it is fixed now. In v1.0 the
        value has been checked at each onkeypress-event, and that was not a
        very good idea in some cases. now the the value will be checked when
        the focus get lost. if the value isn't in the range of min-max, the
        focus can't get lost

        the user can press the ESC key in TFloatEdit to restore the value that
        was set before he made changes


-installation:
copy or extract mpdelver.inc, fledit.pas and fledit.dcr into a folder included
in delphi's search path (like $(DELPHI)\lib). if you have a database aware
version of delphi, you may additionally extract the files dbfloatedit.pas and
dbfloatedit.dcr to install TDBFloatEdit.

select "component"->"install component" in delphi's main menu.

type fledit.pas into the "unit name" edit box.

select "OK".

(db-aware only:) select "add".

(db-aware only:) type dbfloatedit.pas into the "unit name" edit box.

(db-aware only:) select "OK".

after the component package is opened in delphi, click "compile".

after recompiling the package, you should find the components on the
"mirkes.de" component palette page.


-usage:
TFloatEdit, TDBFloatEdit

added published properties:
      Digits: Byte (r/w)
           the number of decimal digits (digits after the decimal separator),
           range 0..18

      Min: Extended (r/w)
           the lowest valid value

      Max: Extended (r/w)
           the highest valid value

      Value: Extended (r/w)
           the current value

      ErrorMessage :string (r/w)
           the message that will be displayed if an invalid (not a number, out
           of range) value is typed in the edit box, if this is the empty
           string ("", default), no error message will be displayed, and the
           value will be reset to a default

      Alignment: TAlignment (r/w)
           how to align the text in the edit box (taLeftJustify, taCenter,
           taRightJustify (default))

      About: string (r)
           shows the version and copyright holder of the component in the
           delphi ide, readonly


added events:
      OnEnterPress : TKeyPressEvent (r/w)
           this event handler is called whenever the ENTER (RETURN) key is
           pressed


changed properties:
      Text: string (r)
           this property is readonly in TFloatEdit, you can just read the
           current value in textual format. to change the value, use the Value
           property instead


TSpinFloatEdit

TSpinFloatEdit inherits all properties of TCustomFloatEdit (as TFloatEdit
does), it additionally implements the following:

added types:
      TComputedIncrementEvent = procedure (Sender: TObject; const bIncrement:
      Boolean) of object;
           this event handler is used in the OnComputedIncrement (see below)
           event


added published properties:
      EditorEnabled: Boolean (rw)
           if this is set to False (default is True), you cannot enter values
           in the edit box, you can just use the increment/decrement buttons
           and keys (up: CURSOR UP, down: CURSOR DOWN)

      Increment: Extended (rw)
           the value that is added to/substracted from the current value on
           increment/decrement event


added events:
      OnComputedIncrement : TComputedIncrement (r/w)
           if a method is assigned to this event handler, it will be called
           whenever an increment/decrement event happens rather than
           adding/substracting the Increment property's value. the boolean
           parameter bIncrement tells you whether an increment (true) or
           decrement (false) event is occured. in this method you may implement
           your own increment/decrement algorithm


-contact:
the author of this package is markus stephany, losheim am see, saarland,
germany.

mailto:vcl[at]mirkes[dot]de (change the obvious!)

http://www.mirkes.de


-license:

(*******************************************************************************

       TFloatEdit, TSpinFloatEdit Copyright 1997-2004 markus stephany

       This package is free software; you can redistribute it and/or modify
       it under the terms of either:

       a) the GNU General Public License as published by the Free
       Software Foundation; either version 2, or (at your option) any
       later version, or

       b) the Clarified Artistic License as used by the authors of NCFTP
       version 3.0.2.

       This package is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
       (Lesser) General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this package in the file named "COPYING"; if not,
       write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
       Boston, MA 02111-1307, USA.

       You should also have received a copy of the Clarified Artistic
       License along with this package in the file named "ARTISTIC"; if not,
       you may find a copy at http://www.mirkes.de/downloads/ARTISTIC.txt

*******************************************************************************)



end.



